#include "modelgrid.h"
#include "datagrid.h"
#include <cmath>
#include "tnt.h"
#include "jama_lu.h"


using namespace std;

double ME_penalty(double x) {
  // Uses an interpolated penalty function similar to that from SS&B
  // (1998) eq 8.  Allows for negative arguments, but gives a pretty stiff penalty.
  if (x > 0.1) {
    return x*log(x);
  } else {
    return (-0.230258-1.30258*(x-0.1));
  }
}


modelgrid::modelgrid() {
  nx=1;
  ny=1;
  psi.resize(5*5);
  kappa.resize(1);
  gamma1.resize(1);
  gamma2.resize(1);
  e1.resize(1);
  e2.resize(1);
  eta=1.;
  b=vector<double>(1,1);
  reg_type="NONE";  
}

modelgrid::modelgrid(int nx_new) {
  // We will not put a buffer on psi.  We know that it is buffered by
  // psi=0, by construction.
  nx=nx_new;
  ny=nx_new;
  n=nx*ny;
  NN=(nx+2)*(ny+2);
  psi.resize((nx+2)*(ny+2));
  kappa.resize(nx*ny);
  gamma1.resize(nx*ny);
  gamma2.resize(nx*ny);
  F1.resize(nx*ny);
  F2.resize(nx*ny);
  e1.resize(nx*ny);
  e2.resize(nx*ny);
  e1_data.resize(nx*ny);
  e2_data.resize(nx*ny);
  
  if (reg_type.find("ME") != string::npos) {
    b=vector<double>(nx*ny,1);
  } else {
    b=vector<double>(nx*ny,0);
  }
  eta=1.;
  
  for (int i=0; i < n ;i++) {
    kappa[i]=0;
    gamma1[i]=0;
    gamma2[i]=0;
    e1[i]=0;
    e2[i]=0;
    e1_data[i]=0;
    e2_data[i]=0;
    F1[i]=0;
    F2[i]=0;
    w.push_back(0);
  }
  
  for (int i=0; i < (nx+2)*(ny+2); i++) {
    psi[i]=0;
  }
  cerr<<NN<<endl;
  create_matrix();
}
double modelgrid::get_data(string name,int i,int j) {
  // If we like, we can say that element -1, and nx yields a value... 0.
  // all names are in uppercase
  int idx=get_index(name,i,j);
  if (name.find("PSI") != string::npos) {
    if (idx != -1) {
      return(psi[idx]);
    } else {
      return 0;
    }
  }
  if (name.find("KAPPA") != string::npos) {
    if (idx != -1) {
      return(kappa[idx]);
    } else {
      return 0;
    }
  }
  if (name.find("GAMMA1") != string::npos) {
    if (idx != -1) {
      return(gamma1[idx]);
    } else {
      return 0;
    }
    
  }
  if (name.find("GAMMA2") != string::npos) {
    if (idx != -1) {
      return(gamma2[idx]);
    } else {
      return 0;
    }
  }

  if (name.find("F1") != string::npos) {
    if (idx != -1) {
      return(F1[idx]);
    } else {
      return 0;
    }
  }
  
  if (name.find("F2") != string::npos) {
    if (idx != -1) {
      return(F2[idx]);
    } else {
      return 0;
    }
  }

  if (name.find("E1") != string::npos) {
    if (idx != -1) {
      return(e1[idx]);
    } else {
      return 0;
    }
  }

  if (name.find("E2") != string::npos) {
    if (idx != -1) {
      return(e2[idx]);
    } else {
      return 0;
    }
  }
  cerr << "Warning! modelgrid.get_data called without correct datatype\n";
  return 0;
}

vector<double> modelgrid::get_data(string name)  {
  if (name.find("PSI") != string::npos) {
    return psi;
  } else if (name.find("KAPPA") != string::npos) {
    return kappa;
  } else if (name.find("GAMMA1") != string::npos) {
    return gamma1;
  } else if (name.find("GAMMA2") != string::npos) {
    return gamma2;
  } else if (name.find("F1") != string::npos) {
    return F1; 
  } else if (name.find("F2") != string::npos) {
    return F2; 
  } else if (name.find("E1") != string::npos) {
    return e1; 
  } else if (name.find("E2") != string::npos) {
    return e2; 
  }
  else {
    cerr << "Type does not exist\n";
    return vector<double>(0);
  }
}

void modelgrid::set_psi(int i,int j,double val)
{
  int idx=get_index("PSI",i,j);
  if (idx != -1) {
    psi[idx]=val;
  } else {
    cerr << i << " , " << j << " outside the domain\n";
  }
  

}

void modelgrid::update_all() {
  update_kappa();
  update_gamma();
  update_ellipticity();
  //update_flexion();
}

void modelgrid::set_psi(vector<double> valarray) {
  if (valarray.size() == psi.size()) {
    psi=valarray;
    
  }
  else {
    cerr << "Input array is not of the correct size\n";\
    return;
  }
  update_all();
}

void modelgrid::update_gamma() {
  // Updates the gamma field based on the current potential field.
  // Generally only called by the codes internal to modelgrid member functions.
  
  double oneoverh2=nx*nx;

  int delta_i_g1[]={0,-1,1,0};
  int delta_j_g1[]={-1,0,0,1};
  double M_1[]={-0.5*oneoverh2,0.5*oneoverh2,0.5*oneoverh2,-0.5*oneoverh2};

  int delta_i_g2[]={-1,1,-1,1};
  int delta_j_g2[]={-1,-1,1,1};
  double M_2[]={0.25*oneoverh2,-0.25*oneoverh2,-0.25*oneoverh2,0.25*oneoverh2};

  int npts=4;

  for (int i=0; i < nx; i++) {
    for (int j=0; j < ny; j++) {
      int idx=get_index("GAMMA1",i,j);
      double newval1=0.;
      double newval2=0.;
      for (int ipt=0; ipt < npts; ipt++) {
	newval1+=M_1[ipt]*get_data("PSI",i+delta_i_g1[ipt],j+delta_j_g1[ipt]);
	newval2+=M_2[ipt]*get_data("PSI",i+delta_i_g2[ipt],j+delta_j_g2[ipt]);
      }
      gamma1[idx]=newval1;
      gamma2[idx]=newval2;
    }
  }
}

void modelgrid::update_kappa() {
  // Updates the kappa field based on the current potential field.
  // Generally only called by the codes internal to modelgrid member functions.
  
  double oneoverh2=nx*nx;

  int delta_i[]={0,-1,0,1,0};
  int delta_j[]={1,0,0,0,-1};
  double M[]={0.5*oneoverh2,0.5*oneoverh2,-2*oneoverh2,0.5*oneoverh2,0.5*oneoverh2};
  int npts=5;

  for (int i=0; i < nx; i++) {
    for (int j=0; j < ny; j++) {
      int idx=get_index("KAPPA",i,j);
      double newval=0.;
      for (int ipt=0; ipt < npts; ipt++) {
	newval+=M[ipt]*get_data("PSI",i+delta_i[ipt],j+delta_j[ipt]);
      }
      kappa[idx]=newval;
    }
  }
}

void modelgrid::update_flexion() {
  // Updates the kappa field based on the current potential field.
  // Generally only called by the codes internal to modelgrid member functions.
  double oneoverh2=nx*nx;
  

  int delta_i_f1[]={2,1,1,1,-2,-1,-1,-1};
  int delta_j_f1[]={0,1,-1,0,0,1,-1,0};
  double M_f1[]={0.25*oneoverh2,0.25*oneoverh2,0.25*oneoverh2,-2*oneoverh2,-0.25*oneoverh2,-0.25*oneoverh2,-0.25*oneoverh2,2*oneoverh2};

  int delta_i_f2[]={1,-1,0,0,1,-1,0,0};
  int delta_j_f2[]={1,1,2,1,-1,-1,-2,-1};
  double M_f2[]={0.25*oneoverh2,0.25*oneoverh2,0.25*oneoverh2,-2*oneoverh2,-0.25*oneoverh2,-0.25*oneoverh2,-0.25*oneoverh2,2*oneoverh2};
  int npts=5;

  for (int i=0; i < nx; i++) {
    for (int j=0; j < ny; j++) {
      int idx=get_index("F1",i,j);
      double newval1=0.;
      double newval2=0.;
      for (int ipt=0; ipt < npts; ipt++) {
	newval1+=M_f1[ipt]*get_data("PSI",i+delta_i_f1[ipt],j+delta_j_f1[ipt]);
	newval2+=M_f2[ipt]*get_data("PSI",i+delta_i_f2[ipt],j+delta_j_f2[ipt]);
      }
      F2[idx]=newval2;
      F1[idx]=newval1;
    }
  }
}

void modelgrid::update_ellipticity() {
  double g1,g2,g;
  for (int i=0; i < nx*ny; i++) {
    g1=gamma1[i]/(1-kappa[i]);
    g2=gamma2[i]/(1-kappa[i]);
    g=sqrt(g1*g1+g2*g2);
    if (g < 1) {
      e1[i]=g1;
      e2[i]=g2;
    } else {
      e1[i]=g1/(g*g);
      e2[i]=g2/(g*g);
    }   
  }
}


// Compute the chi^2 difference between a datagrid and the real data.
double modelgrid::chi2(datagrid d) {
  double chi2_out=0.;
  double regval;
  double kappaave=0; // This needs to be positive!
  for (int i=0; i < nx ; i++) {
    for (int j=0; j < ny; j++) {
      chi2_out+=d.get_data("W",i,j)*pow((get_data("E1",i,j)-d.get_data("E1",i,j)),2);
      chi2_out+=d.get_data("W",i,j)*pow((get_data("E2",i,j)-d.get_data("E2",i,j)),2);
      if (reg_type.find("ME") != string::npos) {
	kappaave+=kappa[i*ny+j]/(nx*ny);
      }
    }
  }
  if (kappaave < 0.1) {
    kappaave=0.1;
  }
  
  if (reg_type.find("ME") != string::npos) {
    //cerr << "Using Maximum Entropy Regularization: " << chi2_out << "\n";
    //cerr << "  " << ME_penalty(0.) << " " << eta << " " << b << "\n";
    double x=0;
    for (int i=0; i < nx*ny; i++) {
      x=kappa[i]/b[i];//kappaave;
      //cerr << x << "\n";
      chi2_out+=eta*b[i]*ME_penalty(x);
    }
    //cerr << "   " << chi2_out << "\n";
  } else if (reg_type.find("BRADAC") != string::npos) {
    for (int i=0; i < nx*ny; i++) {
      chi2_out+=eta*pow((kappa[i]-b[i]),2);
    }
  }
  return chi2_out;

}

int modelgrid::get_index(string name, int i, int j) {
  if (name.find("PSI") != string::npos) {
    if ((i >= -1)&&(i < nx+1)&&(j>=-1)&&(j < ny+1)) {
      return((i+1)*(ny+2)+j+1);
    } else {
      return -1;
    }
  }
  else {
    if ((i >=0 )&&(i < nx)&&(j >=0)&&(j <ny)) {
      return(i*ny+j);
    }
    else {
      return -1;
    }
  }
}

modelgrid modelgrid::refine(int ismooth) {
  // Takes an old grid, and refines it by a factor of two.  Psi is
  // interpolated as a good starting point, and the b-values are set
  // as the kappa/kappa_ave from the previous grid.

  // ismooth=0 (default) gives nearest gridcell interpolation
  // ismooth=1 gives CIC interpolation

  update_all();
  int nx_new=nx*2,ny_new=ny*2;
  vector<double> b_new(nx_new*ny_new);
  modelgrid outgrid(nx_new);
  double kappa_ave=0;
  // Inherit the regularization type and eta from this data.
  outgrid.set_reg_type(reg_type);
  outgrid.set_eta(eta);
  // Compute kappa_ave, and
  if (reg_type.find("ME") != string::npos) {
    for (int i=0; i < nx; i++) {
      for (int j=0; j < ny; j++) {
	kappa_ave+=get_data("KAPPA",i,j)/(nx*ny);
      }
    }
  }
  // Set b to the kappa/kappa_ave from the previous grid.
  if ((reg_type.find("ME") != string::npos) || (reg_type.find("BRADAC") != string::npos)) {
    if (ismooth == 0) {
      for (int i=0; i < nx; i++) {
	for (int j=0; j < ny; j++) {
	  double b=get_data("KAPPA",i,j);///kappa_ave;
	  for (int i1=2*i; i1< 2*i+2; i1++) {
	    for (int j1=2*j; j1 < 2*j+2; j1++) {
	      b_new[i1*ny_new+j1]=b;
	    }
	  }
	}
      }
    }
    else if (ismooth == 1) {
      // Note -- don't use these definitions of x positions anywhere else!
      // We're just doing it for convenience
      for (int i1=0; i1 < nx_new; i1++) {
	double x1=i1/2.-0.25;
	int i0=floor(x1);
	double wx=x1-i0;
	for (int j1=0; j1 < ny_new; j1++) {
	  double y1=j1/2.-0.25;
	  int j0=floor(y1);
	  double wy=y1-j0;
	  	  
	  b_new[i1*ny_new+j1]=(1-wx)*(1-wy)*get_data("KAPPA",i0,j0)+
	    (1-wx)*wy*get_data("KAPPA",i0,j0+1)+
	    wx*(1-wy)*get_data("KAPPA",i0+1,j0)+
	    wx*wy*get_data("KAPPA",i0+1,j0+1);
	}
      }
    }
  }
  outgrid.set_b(b_new);
  // Finally, do the edges.
  
  outgrid.fit_b();

  outgrid.update_all();
  return(outgrid);
}


void modelgrid::create_matrix(){
  int idx1,idx2;
  vector<double> dum(NN,0);
  cerr<<"i am here"<<endl;
  for (int i=0; i < NN; i++) {
     K_arr.push_back(dum);
     G1_arr.push_back(dum);
     G2_arr.push_back(dum);
  }
  
  double norm=0.5*NN;
  for (int i1=0; i1 < (nx+2); i1++) {
    for (int j1=0; j1 < (ny+2); j1++) {
      idx1=i1*(ny+2)+j1;
      // w=1 delta_i=0 delta_j=+1
      // w=-1 delta_i=0 delta_j=+1
       
      if (j1 < (ny+2)-1) {
	idx2=i1*(ny+2)+j1+1;
	K_arr[idx1][idx2]=norm;
	G1_arr[idx1][idx2]=-norm;
      }
      // w=1 delta_i=-1 delta_j=0
      if (i1 > 0) {
	idx2=(i1-1)*(ny+2)+j1;
	K_arr[idx1][idx2]=norm;
	G1_arr[idx1][idx2]=norm;
      }
      // w=-4 delta_i=0 delta_j=0
      K_arr[idx1][idx1]=-4.*norm;
      // w=1 delta_i=1 delta_j=0
      if (i1 < (nx+2)-1) {
	idx2=(i1+1)*(ny+2)+j1;
	K_arr[idx1][idx2]=norm;
	G1_arr[idx1][idx2]=norm;
      }
      // w=1 delta_i=0 delta_j=-1
      if (j1 >0) {
	idx2=i1*(ny+2)+j1-1;
	K_arr[idx1][idx2]=norm;
	G1_arr[idx1][idx2]=-norm;
      }
      
      //w=0.5 delta_i=-1 delta_j=-1
      if (j1 >0 && i1 >0) {
	idx2=(i1-1)*(ny+2)+j1-1;
	G2_arr[idx1][idx2]=0.5*norm;
      }
      //w=-0.5 delta_i=-1 delta_j=+1
      if (j1 <(ny+2)-1 && i1 >0) {
	idx2=(i1-1)*(ny+2)+j1+1;
	G2_arr[idx1][idx2]=-0.5*norm;
      }
      //w=-0.5 delta_i=+1 delta_j=-1
      if (j1 >0 && i1<(nx+2)-1 ) {
	idx2=(i1+1)*(ny+2)+j1-1;
	G2_arr[idx1][idx2]=-0.5*norm;
      }
      //w=0.5 delta_i=+1 delta_j=+1
      if (j1 <(ny+2)-1 && i1<(nx+2)-1 ) {
	idx2=(i1+1)*(ny+2)+j1+1;
	G2_arr[idx1][idx2]=0.5*norm;
      }
    }
  }
 
}

// fit_b()
// This function fits the kappa field of a grid to the values set by the
// b array, subsequently setting psi.
void modelgrid::fit_b() {
 
  // Note: by construction, the psi's on the edges of the grid will be zero.
  
  // Recall A_ij psi_j=b_i

  int idx2;
  double norm=0.5*n;
  Array2D<double> A_arr(n,n);
  Array1D<double> b_arr(n),psi_new(n);
  for(int i=0;i<n;i++){
    b_arr[i]=0;
    for(int j=0;j<n;j++){
      A_arr[i][j]=0;
    }
  }

  for (int i1=0; i1 < (nx); i1++) {
    for (int j1=0; j1 < ny; j1++) {
      int idx1=i1*ny+j1;
      b_arr[idx1]=b[idx1];
// w=1 delta_i=0 delta_j=+1
      if (j1 < ny-1) {
        idx2=i1*ny+j1+1;
        A_arr[idx1][idx2]=norm;
      }
      // w=1 delta_i=-1 delta_j=0
      if (i1 > 0) {
        idx2=(i1-1)*ny+j1;
        A_arr[idx1][idx2]=norm;
      }
      // w=-4 delta_i=0 delta_j=0
      A_arr[idx1][idx1]=-4.*norm;
      // w=1 delta_i=1 delta_j=0
      if (i1 < nx-1) {
        idx2=(i1+1)*ny+j1;
        A_arr[idx1][idx2]=norm;
      }
      // w=1 delta_i=0 delta_j=-1
      if (j1 >0) {
        idx2=i1*ny+j1-1;
        A_arr[idx1][idx2]=norm;
      }
      
    }
  }   
  
  JAMA::LU<double> A_lu(A_arr);
  psi_new=A_lu.solve(b_arr);
  for (int i=0; i < nx; i++) {
    for (int j=0; j < ny; j++) {
      set_psi(i,j,psi_new[i*ny+j]);
    }
  }
  update_all();
}

vector<double> modelgrid::psi_newton_grid()
{
  int i,idx,j;
  Array1D<double> Q(NN),dpsi(NN);
  Array2D<double> M(NN,NN);
  Array2D<double> A(3*n,NN);//number of constraints=3*n;  //e1,e2,kappa
  vector<double> dyarr(3*n,0);
  vector<double> dpsi_ret(NN,0);
  double epsilon=0.000001;
  double de1dgamma1,de1dgamma2,de1dkappa,de2dgamma1,de2dgamma2,de2dkappa;
  
  double dum;
  for( i=0;i<3*n;i++){
    for(int j=0;j<NN;j++){
      A[i][j]=0;
    }
  }

  for( j=0;j<NN;j++){
    Q[j]=0;
    dpsi[j]=0;
    for(int k=0;k<NN;k++){
     M[j][k]=0.;
    }
  }
  //cerr << "I am\n";
 
  for(int p=0;p<nx;p++){
    for(int q=0;q<ny;q++){
      i=p*ny+q;
    //Derivatives of  ellipticities
	dum=fabs(1-kappa[i])-sqrt(gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]);
	if (dum >= 0){
	  de1dgamma1=1./(1-kappa[i]);
	  de1dgamma2=0.;
	  de1dkappa=gamma1[i]/((1-kappa[i])*(1-kappa[i]));
	  de2dgamma1=0.;
	  de2dgamma2=1./(1-kappa[i]);
	  de2dkappa=gamma2[i]/((1-kappa[i])*(1-kappa[i]));
      
	} else{ 
	  
	  de1dgamma1=(kappa[i]-1.)*(gamma1[i]*gamma1[i]-gamma2[i]*gamma2[i])/pow((gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6),2);
	  de1dgamma2=-2.*(1-kappa[i])*gamma1[i]*gamma2[i]/pow((gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6),2);
	  de1dkappa=-gamma1[i]/(gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6);
	  
	  de2dgamma1=-2*(1-kappa[i])*gamma1[i]*gamma2[i]/pow((gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6),2);
	  de2dgamma2=(1-kappa[i])*(gamma1[i]*gamma1[i]-gamma2[i]*gamma2[i])/pow((gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6),2);
	  de2dkappa=-gamma2[i]/(gamma1[i]*gamma1[i]+gamma2[i]*gamma2[i]+1e-6);
	}
	
	dyarr[i]=e1_data[i]-e1[i];
	dyarr[i+n]=e2_data[i]-e2[i];
	dyarr[i+2*n]=b[i]-kappa[i];
	//cerr<<b[i]<<endl;
	idx=(p+1)*(ny+2)+q+1;
	for(int l=0;l<nx+2;l++){
	  for(int m=0;m<ny+2;m++){
	    j=l*(ny+2)+m;
	    
	    //cerr<<"I AM HERE  "<<j<<endl;
	    A[i][j]=K_arr[idx][j]*de1dkappa+G1_arr[idx][j]*de1dgamma1+G2_arr[idx][j]*de1dgamma2;
	    A[i+n][j]=K_arr[idx][j]*de2dkappa+G1_arr[idx][j]*de2dgamma1+G2_arr[idx][j]*de2dgamma2;
	    A[i+2*n][j]=K_arr[idx][j];
	    Q[j]+=w[i]*A[i][j]*dyarr[i]+w[i]*A[i+n][j]*dyarr[i+n]+eta*A[i+2*n][j]*dyarr[i+2*n];
	    for(int k=0;k<=j;k++){
	      M[j][k]+=w[i]*A[i][j]*A[i][k]+w[i]*A[i+n][j]*A[i+n][k]+eta*A[i+2*n][j]*A[i+2*n][k];
	      M[k][j]=M[j][k];
	      if (j == k)  M[j][k]+=1.e-5;
	    }
	    //cerr<<"I AM HERE  "<<j<<endl;
	  }
	}
    }
  }
 
  JAMA::LU<double> A_lu(M);
  dpsi=A_lu.solve(Q);
  //cerr<<dpsi[0]<<"  "<<dpsi[nx*ny]<<"  "<<dpsi[nx]<<"  "<<dpsi[nx*ny+nx]<<endl;;
  
  for(int i=0;i<NN;i++){
    dpsi_ret[i]=dpsi[i];
  }
 
  return dpsi_ret;
}


void modelgrid::set_data(datagrid d){
  e1_data=d.get_data("E1");
  e2_data=d.get_data("E2");
  w=d.get_data("W"); // This needs to Qe taken care of... put in datapbl
}

vector<vector<double> > modelgrid::get_arr(string name){
  if(name == "G1_arr") return G1_arr;
  if(name == "G2_arr") return G2_arr;
  if(name == "K_arr") return K_arr;
}


