#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include "datagrid.h"
#include "modelgrid.h"
#include "psi_iterate.h"
#include "binnedgrid.h"



using namespace std;

int main(int argc, char* argv[]) {
  // This code will take a configuration file of the form:
  // 1. datafile (containing column delimited x y e1 e2 sigma_e)
  // 2. int constsigma (0 to use the parameters in the datafile, 
  //                    1 to use the parameter in line 3)
  // 3. double sigma_e (if constsigma=0, then this number is thrown away)
  // 4. Regularlization method ("ME", "NONE") are all that are supported.
  // 5. eta (if no regularization, leave blank
  // 5a. Number of iterations per refinement
  // 6. nx_initial 
  // 7. # of refinements, so that nx_final=nx_initial*2^n
  // 8. n_iso // Number of isothermal spheres for an initial fit
  // 9-11,12-14... [x_iso,y_iso,theta_e_iso]
  if (argc == 1) {
    cerr << "calling procedure is: recon_grid [configfile] \n";
    return(0);
  } 
  // First, read in the configuration file
  
  fstream datafile, configfile;
  configfile.open(argv[1],ios::in);
  if (! configfile.is_open()) {
    cerr << "Boo!  File does not exist.\n";
    return(0);
  }
  char dataname[200];
  string reg_method;
  double eta;
  vector<double> x_arr ,y_arr,w_arr,e1_arr,e2_arr,psi0_arr,ktrue_arr;
  configfile >> dataname;
  datafile.open(dataname,ios::in);
  int constsigma,nx,itermax;
  double sige;
  int nrefine;
  configfile >> constsigma;
  configfile >> sige;
  configfile >> reg_method;
  configfile >> eta;
  configfile >> itermax;
  configfile >> nx;
  configfile >> nrefine;


  // now read in the data;
  int n=0;
  string line;
  double d1,d2,d3,d4,d5,d6,d7,d8;
  while  (! (datafile.eof() || datafile.peek()== EOF)) {
    datafile >> d1 >> d2 >> d3 >> d4 >> d5;
    x_arr.push_back(d1);
    y_arr.push_back(d2);
    e1_arr.push_back(d3);
    e2_arr.push_back(d4);
    ktrue_arr.push_back(d5);
    w_arr.push_back(100.);
    if (constsigma == 1) {
      w_arr[n]=1./(sige*sige);
    } 
    getline(datafile,line);
    n++;
  }
  datafile.close();
  
  cerr << "n=" << n << "\n";
  // Now, create a new datagrid object
  modelgrid mymodel(nx);
  mymodel.set_reg_type(reg_method);
  mymodel.set_eta(eta);
  int ny=mymodel.get_ny();
  datagrid mydata(nx,n,x_arr,y_arr,e1_arr,e2_arr,w_arr);
  mymodel.set_data(mydata);

  // Set up the initial conditions (if any)
  int n_iso;
  double xc,yc,te;
  configfile >> n_iso;
  vector<double>psi_new;
  for (int i=0; i < (nx+2)*(ny+2); i++) {
    psi_new.push_back(i);
  }
  for (int i_iso=0; i_iso < n_iso; i_iso++) {
    configfile >> xc;
    configfile >> yc;
    configfile >> te;
    for (int i=-1; i < nx+1; i++) {
      for (int j=-1; j < ny+1; j++) {
	double theta=sqrt(pow((double(i+0.5)/nx-xc),2)+pow((double(j+0.5)/ny-yc),2));
	psi_new[(i+1)*(ny+2)+(j+1)]+=te*theta;
      }
    }
  }
  mymodel.set_psi(psi_new);


  /****************************************************************************

     Above this point is all about reading in data, setting parameters, 
     and adjusting initial conditions.  Below this point, we do the 
     actual minimization
  ****************************************************************************/


  int q=0;
  int NN=(nx+2)*(ny+2);
  double dof=2*nx*ny-NN;
    
  vector<double> b;
  cerr << "  preliminary chi^2= " << mymodel.chi2(mydata)/dof << "\n";

  
  for (int irefine=0; irefine < nrefine+1; irefine++) {
    cerr << "Iteration :" << irefine << "\n";
    cerr << "  nx = " << nx << "\n";
    // set the data
    mydata=datagrid(nx,n,x_arr,y_arr,e1_arr,e2_arr,w_arr);
    mymodel.set_data(mydata);
    vector<double> psi_new=mymodel.get_data("PSI");
    cerr << "  preliminary chi^2= " << mymodel.chi2(mydata) << "\n";
    //itermax=itermax/(irefine+1); 
    double factor=0.6;
    if (irefine > 0){
      eta=eta/irefine;
      mymodel.set_eta(eta);
    }

    double dchi=1000.;
    double eps=0.1;
    double chi=mymodel.chi2(mydata);
    double chi0;
    while (dchi > eps){
      cerr << "I am inside iter\n";
      vector<double> dpsi=mymodel.psi_newton_grid();//psi_iterate(mymodel,mydata);//
      vector<double> psi_new=mymodel.get_data("PSI");
      for (int i=0; i < NN; i++) {
	psi_new[i]+=factor*dpsi[i];
	}
	mymodel.set_psi(psi_new);
      cerr << "  chi2 after iteration = " << mymodel.chi2(mydata) << "\n";
      chi0=mymodel.chi2(mydata);
      dchi=fabs(chi-chi0);
      chi=chi0;
    }
    //eta=eta/2.;
    //mymodel.set_eta(eta);
    // Check to see if we're doing another iteration!
    
    if (irefine != nrefine) {
      if (irefine!=0) q=1;
      mymodel=mymodel.refine(1);
      b=mymodel.get_b();
      nx=nx*2;
      ny=ny*2;
      NN=(nx+2)*(ny+2);
      dof=2*nx*ny-NN;
      mydata=datagrid(nx,n,x_arr,y_arr,e1_arr,e2_arr,w_arr);
      //itermax=0;
    }
 }
  mymodel.set_eta(0);
  cerr << "  chi2 after iteration (without eta) =" << mymodel.chi2(mydata)/dof << "\n";
  
  binnedgrid ktruebin(x_arr,y_arr,ktrue_arr,nx,nx);
  vector<double> ktrue_bin=ktruebin.get_data();
  vector<int> n_grid=ktruebin.get_count();
  //Figure out dof by looking at number of cells that have data
  int num=0;
  for (int i=0;i<nx*ny;i++){
    if (n_grid[i] > 0) num+=1;
  }

  dof=num;
  cerr<<"dof=  "<<dof<<endl;
    

  // Now return various goodness measures.
  
  // 1 Compute the nearest gridcell and compute delta kappa^2
  double stat1=0;
  
  // We need to optimize for lambd
 
  double Sxx=0,Sxy=0,Sy=0,Sx=0;
  for (int i=0; i < n; i++) {
    int i0=int(x_arr[i]*nx);
    int j0=int(y_arr[i]*ny);
    double x=mymodel.get_data("KAPPA",i0,j0);
    double y=ktrue_arr[i];
    //cout << x << " " << y << "\n";
    Sxx+=x*x;
    Sxy+=x*y;
    Sx+=x;
    Sy+=y;
    //n_grid[i0*nx+j0]+=1;
  }
  double lambda=(Sxy-Sy-Sx+n)/(Sxx-2*Sx+n);

  
  for (int i=0; i < n; i++) {
    int i0=int(x_arr[i]*nx);
    int j0=int(y_arr[i]*ny);
    double x=mymodel.get_data("KAPPA",i0,j0);
    double y=ktrue_arr[i];    
    stat1+=pow(y-(lambda*x+(1-lambda)),2)/n;
    cout<<x_arr[i]<<"  "<<y_arr[i]<<"  "<<lambda*x+(1-lambda)<<"  "<<ktrue_arr[i]<<endl;
  }

  double stat2=0;
  double stat3=0;
  Sxx=0;Sxy=0;Sy=0;Sx=0;
  for (int i=0; i < nx; i++) {
    for (int j=0; j < nx; j++) {
      double x=mymodel.get_data("KAPPA",i,j);
      double y=ktrue_bin[i*nx+j];
      Sxx+=x*x;
      Sxy+=x*y;
      Sx+=x;
      Sy+=y;
    }
  }

  double lambda_grid=(Sxy-Sy-Sx+nx*nx)/(Sxx-2*Sx+nx*nx);
  cerr<<lambda_grid<<" "<<lambda<<endl;
  int c=0;
  for (int i=0; i < nx; i++) {
    for (int j=0; j < nx; j++) {
      double x=mymodel.get_data("KAPPA",i,j);
      double y=ktrue_bin[i*nx+j];
      if (ktrue_bin[i*nx+j] !=0 ){
	stat2+=pow((y-(lambda_grid*x+(1-lambda_grid))),2);
	c+=1;
      }
      //cout<< i << "  " << j << "  " <<(y-(lambda_grid*x+(1-lambda_grid)))<<"  "<<(lambda_grid*x+(1-lambda_grid))<<"  "<<y<<endl;
      
      stat3+=pow((y-(lambda_grid*x+(1-lambda_grid))),2)*n_grid[i*nx+j]/n;
    }
  }
  cerr<<stat2<<endl;
  stat2=stat2/c;
  cerr<<c<<endl;
  cerr << "Awesomeness statistic1 = " << stat1 << "\n";
  cerr << "Awesomeness statistic2 = " << stat2 << "\n";
  cerr << "Awesomeness statistic3 = " << stat3 << "\n";
  //mymodel=f.get_model();

  vector<double> kappa=mymodel.get_data("KAPPA");
  vector<double> gamma1=mymodel.get_data("GAMMA1");
  vector<double> gamma2=mymodel.get_data("GAMMA2");

  for (int i=0; i < nx; i++) {
    for (int j=0; j < nx; j++) {
      // cout << i << " " << j << " " << mymodel.get_data("KAPPA",i,j)<< " " 
//	   << mymodel.get_data("E1",i,j)<< "" << mymodel.get_data("E2",i,j)<< " " 
//	   << mydata.get_data("E1",i,j) << " "<< mydata.get_data("E2",i,j) << " " 
//	   << mydata.get_data("W",i,j) <<  "\n";
      // cout<< i << "  " << j << "  " <<(mymodel.get_data("KAPPA",i,j)*lambda_grid+(1-lambda_grid))<<"  "<<ktrue_bin[i*nx+j]<<endl;
      
    }
  }





  return 0;
}
